luafontcomp        = { }

luafontcomp.module = {
   name         = "luafontcomp",
   version      = 0.05,
   date         = "2011/03/16",
   description  = "Selecting complementary fonts for LuaLaTeX.",
   author       = "Kazuki Maeda",
   copyright    = "Kazuki Maeda",
   license      = "MIT License"
}

local error, warning, info, log = luatexbase.provides_module(luafontcomp.module)

utf = unicode.utf8

local prefix_name = "name:"
local prefix_file = "file:"
local prefix_vf = "vf:"

local dummy_fonttable = {
   name = "dummy",
   characters = {
      [utf.byte(" ")] = {
         width = 655360
      }
   },
   fonts = {
      {name = 'cmr10'}
   },
   type = "virtual"
}

local vf_map_table = {}          -- vf_map_table[vfbasename]...
local vf_id_table = {}           -- vf_id_table[vfid]...
local vf_name_table = {}         -- vf_name_table[vfname][size] -> fontid
local rf_cache_table = {}        -- rf_cache_table[rfname]...

-- vfname must be "name:vf:..." or "vf:..."
local function fontname_split(name)
   local prefix = ""
   local basename = name
   local features = ""
   local shape = ""

   local vf_flag = false
   
   local tmp = utf.sub(basename, 1, utf.len(prefix_name))
   
   if tmp == prefix_name or tmp == prefix_file then
      prefix = tmp
      basename = utf.sub(basename, utf.len(prefix_name)+1)
   end

   if utf.sub(basename, 0, utf.len(prefix_vf)) == prefix_vf then
      prefix = prefix .. prefix_vf
      basename = utf.sub(basename, utf.len(prefix_vf)+1)
      vf_flag = true
   end
   
   local p = utf.find(basename, ":")
   if p then
      features = utf.sub(basename, p+1)
      basename = utf.sub(basename, 1, p-1)
   end

   if vf_flag then
      p = utf.find(basename, "/")
      if p then
         shape = utf.sub(basename, p+1)
         basename = utf.sub(basename, 1, p-1)
      end
   end
   
   return prefix, basename, features, shape
end

local function get_decimal(repr)
   if utf.find(repr, '"') == 1 then
      return tonumber(utf.sub(repr, 2), 16)
   else
      return tonumber(repr)
   end
end

local function create_range_table(range)
   local error_msg = 'parse error: "range=' .. range .. '"'
   local range_table = {}
   local range_min = 0
   local range_max = 0x10FFFF

   -- `<char> -> utf.byte(<char>)
   local tmp = ''
   repeat
      local p = utf.find(range, '`')
      if not p then break end
      tmp = tmp .. utf.sub(range, 1, p-1) .. utf.byte(utf.sub(range, p+1, p+1))
      range = utf.sub(range, p+2)
   until false
   range = tmp .. range
   repeat
      local p = utf.find(range, ',')
      local sub_range
      if not p then 
         sub_range = range
         range = ''
      else
         sub_range = utf.sub(range, 1, p-1)
         range = utf.sub(range, p+1)
      end

      if sub_range ~= '' then
         p = utf.find(sub_range, '-')
         if not p then 
            tmp = get_decimal(sub_range)
            if not tmp then error(error_msg) end
            table.insert(range_table, {from = tmp, to = tmp})
         else
            local from = utf.sub(sub_range, 1, p-1)
            local to = utf.sub(sub_range, p+1)
            if from == '' then 
               from = range_min
            else
               from = get_decimal(from)
            end
            if to == '' then 
               to = range_max
            else
               to = get_decimal(to)
            end               
            if not from or not to then error(error_msg) end
            table.insert(range_table, {from = from, to = to})
         end
      end
   until range == ''
   return range_table
end

local function check_range(range_table, val)
   local flag = false
   for k, v in pairs(range_table) do
      if v.from <= val and val <= v.to then 
         flag = true
         break
      end
   end
   return flag
end

-- user interface
function luafontcomp.set_vf_map_table(rfname, scale, vfname, range, exclude, overwrite)
   -- check vfname
   local vfbasename
   , vfbasename = fontname_split(vfname)
   if vfbasename ~= vfname then error('invalid vf name: "' .. vfname .. '"') end

   if not vf_map_table[vfname] then
      vf_map_table[vfname] = {}
      vf_map_table[vfname].fonts = {}
      vf_map_table[vfname].characters = {}
   end

   -- set the real font entry if necessary
   local rfnum = nil
   for k, v in pairs(vf_map_table[vfname].fonts) do
      if v[k] and v[k].name == rfname then rfnum = k; break; end
   end
   if not rfnum then
      table.insert(vf_map_table[vfname].fonts, {name = rfname, scale = scale})
      rfnum = table.maxn(vf_map_table[vfname].fonts)
   end

   -- load the real font if necessary
   local rfprefix, rfbasename
   rfprefix, rfbasename = fontname_split(rfname)
   rfbasename = rfprefix .. rfbasename
   if not rf_cache_table[rfbasename] then
      local fontdata = fonts.define.read(rfbasename)
      if type(fontdata) ~= "table" then
         fontdata = fonts.ids[fontdata]
      end
      rf_cache_table[rfbasename] = {}
      rf_cache_table[rfbasename].characters = fontdata.characters
      rf_cache_table[rfbasename].parameters = fontdata.parameters
   end

   -- parameters
   if not vf_map_table[vfname].parameters then
      vf_map_table[vfname].parameters = {}
      for k, v in pairs(rf_cache_table[rfbasename].parameters) do
         vf_map_table[vfname].parameters[k] = rf_cache_table[rfbasename].parameters[k] * scale
      end
   end

   -- characters
   local range_table = create_range_table(range)
   local exclude_table = create_range_table(exclude)

   for k, v in pairs(rf_cache_table[rfbasename].characters) do
      if check_range(range_table, k) and not check_range(exclude_table, k) and 
         (not vf_map_table[vfname].characters[k] or overwrite) then
         vf_map_table[vfname].characters[k] = rfnum
      end
   end

end

local function copy_features_table(srcid, destid)
   if not fonts.ids[srcid].shared then return end
   for k1, v1 in pairs(fonts.ids[srcid].shared.otfdata.luatex.features) do
      if not fonts.ids[destid].shared.otfdata.luatex.features[k1] then
         fonts.ids[destid].shared.otfdata.luatex.features[k1] = {}
      end
      for k2, v2 in pairs(fonts.ids[srcid].shared.otfdata.luatex.features[k1]) do
         if not fonts.ids[destid].shared.otfdata.luatex.features[k1][k2] then
            fonts.ids[destid].shared.otfdata.luatex.features[k1][k2] = {}
         end
         for k3, v3 in pairs(fonts.ids[srcid].shared.otfdata.luatex.features[k1][k2]) do
            if not fonts.ids[destid].shared.otfdata.luatex.features[k1][k2][k3] then
               fonts.ids[destid].shared.otfdata.luatex.features[k1][k2][k3] = {}
            end
            for k4, v4 in pairs(fonts.ids[srcid].shared.otfdata.luatex.features[k1][k2][k3]) do
               fonts.ids[destid].shared.otfdata.luatex.features[k1][k2][k3][k4] = true                        
            end
         end
      end
   end
end

-- define a virtual font
local function define_vf(name, vfsize, vfid)
   if vfsize and vfsize < 0 then vfsize = -vfsize * 655.36 end
   local vfprefix, vfbasename, vffeatures, vfshape
   vfprefix, vfbasename, vffeatures, vfshape = fontname_split(name)
   if vfprefix == prefix_vf or vfprefix == prefix_name .. prefix_vf then
      vf_id_table[vfid] = {}
      vf_id_table[vfid].basename = vfbasename
      vf_id_table[vfid].fonts = {}

      local vfname = vfbasename
      if vfshape ~= "" then vfname = vfname .. "/" .. vfshape end
      if vffeatures ~= "" then vfname = vfname .. ":" .. vffeatures end
      
      if not vf_map_table[vfbasename] then
         error('virtual font "'.. vfbasename .. '" is not defined.')
      end

      if not vf_name_table[vfname] then
         vf_name_table[vfname] = {}
      end
      
      -- check if already defined
      if vf_name_table[vfname][vfsize] then
         return vf_name_table[vfname][vfsize]
      end

      local convert_table = {}
      fonts.ids[vfid] = {shared = {otfdata = {luatex = {features = {}}}}} 

      for k, v in pairs(vf_map_table[vfbasename].fonts) do
         -- prepare corresponding real fonts
         local rfname_orig = v.name
         local rfscale = v.scale
         local rfprefix, rfbasename, rffeatures
         rfprefix, rfbasename, rffeatures = fontname_split(rfname_orig)
         
         local rfname = rfprefix .. rfbasename 
         if vfshape ~= "" and rfprefix ~= prefix_file then 
            rfname = rfname .. "/" .. vfshape 
         end
         local features = rffeatures
         if rffeatures ~= "" and vffeatures ~= "" then 
            features = features .. ";"
         end
         features = features .. vffeatures
         if features ~= "" then rfname = rfname .. ":" .. features end
         local rfid = font.nextid()
         local rfdata = fonts.define.read_orig(rfname, vfsize*rfscale, rfid)
         if type(rfdata) == "table" then
            rfid = font.define(rfdata)
         else
            rfid = rfdata
         end
         vf_id_table[vfid].fonts[k] = rfid
         copy_features_table(rfid, vfid)
      end

      -- construct the virtual font table
      local fonttable = dummy_fonttable
      fonttable.name = vfname
      fonttable.size = vfsize
      fonttable.fonts = {{name = vf_map_table[vfbasename].fonts[1].name}}
      fonttable.parameters = {}
      local scale = vfsize/655360
      for k, v in pairs(vf_map_table[vfbasename].parameters) do
         fonttable.parameters[k] = vf_map_table[vfbasename].parameters[k] * scale
      end

      vf_name_table[vfname][vfsize] = vfid            

      fonts.ids[vfid].fullname = vfname
      
      return fonttable
   end
end

-- redefine fonts.define.read (otfl-font-def.lua)
fonts.define.read_orig = fonts.define.read
function fonts.define.read(name, size, id)
   local fontdata = define_vf(name, size, id)
   if not fontdata then fontdata = fonts.define.read_orig(name, size, id) end
   return fontdata
end

-- filter
local function node_filter(head, ...)
   local cn = head
   repeat
      if node.type(cn.id) == "glyph" and vf_id_table[cn.font] then
         local vfbasename = vf_id_table[cn.font].basename
         local vffontnum = vf_map_table[vfbasename].characters[cn.char]
         if vffontnum then
            cn.font = vf_id_table[cn.font].fonts[vffontnum]
         end
      end
      cn = cn.next
   until not cn
   return head
end

-- priority of node_filter must be smaller than luaotfload's one
luatexbase.add_to_callback('pre_linebreak_filter',
                            node_filter,
                           'luafontcomp.pre_linebreak_filter',
                           luatexbase.priority_in_callback(
                              'pre_linebreak_filter',
                              'luaotfload.pre_linebreak_filter'))
luatexbase.add_to_callback('hpack_filter',
                            node_filter,
                           'luafontcomp.hpack_filter',
                           luatexbase.priority_in_callback(
                              'hpack_filter',
                              'luaotfload.hpack_filter'))
